#define _GNU_SOURCE
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gen.h"
#include "grep_filter.h"
#include "utils.h"

grep_filter_t *gf_list = NULL;
int n_gf_list = 0;

BOOL add_grep_filter(const char *re, const char *sserver, const char *schannel, const char *file, char **err)
{
	int rc = 0;

	gf_list = (grep_filter_t *)realloc(gf_list, (n_gf_list + 1)* sizeof(grep_filter_t));
	memset(&gf_list[n_gf_list], 0x00, sizeof(grep_filter_t));

	rc = regcomp(&gf_list[n_gf_list].filter, re, REG_NOSUB);
	if (rc)
	{
		char buffer[4096] = { 0 };

		regerror(rc, &gf_list[n_gf_list].filter, buffer, sizeof buffer);
		*err = strdup(buffer);

		return FALSE;
	}

	gf_list[n_gf_list].re = strdup(re);
	gf_list[n_gf_list].sserver = strlen(sserver) ? strdup(sserver) : NULL;
	gf_list[n_gf_list].schannel = strlen(schannel) ? strdup(schannel) : NULL;
	gf_list[n_gf_list].file = strdup(file);
	n_gf_list++;

	*err = NULL;

	return TRUE;
}

BOOL process_grep_filter(const char *sserver, const char *schannel, const char *line, char **err)
{
	BOOL ok = TRUE;
	int loop = 0;

	*err = NULL;

	for(loop=0; loop<n_gf_list && ok; loop++)
	{
		if ((gf_list[loop].sserver == NULL || strcasecmp(gf_list[loop].sserver, sserver) == 0) &&
			(gf_list[loop].schannel == NULL || strcasecmp(gf_list[loop].schannel, schannel) == 0))
		{
			if (regexec(&gf_list[loop].filter, line, 0, NULL, 0) == 0)
			{
				FILE *fh = fopen(gf_list[loop].file, "a+");
				if (!fh)
				{
					asprintf(err, "Cannot open %s: %s\n", gf_list[loop].file, strerror(errno));
					ok = FALSE;
				}
				else
				{
					if (fprintf(fh, "%s\n", line) <= 0)
					{
						asprintf(err, "Cannot write to %s: %s\n", gf_list[loop].file, strerror(errno));
						ok = FALSE;
					}

					fclose(fh);
				}
			}
		}
	}

	return ok;
}

void free_grep_filters(void)
{
	int loop = 0;

	for(loop=0; loop<n_gf_list; loop++)
	{
		regfree(&gf_list[loop].filter);
		myfree(gf_list[loop].re);
		myfree(gf_list[loop].sserver);
		myfree(gf_list[loop].schannel);
		myfree(gf_list[loop].file);
	}

	free(gf_list);
	gf_list = NULL;

	n_gf_list = 0;
}

BOOL dump_grep_filters(FILE *fh)
{
	int loop = 0;

	if (n_gf_list)
		fprintf(fh, ";\n");

	for(loop=0; loop<n_gf_list; loop++)
	{
		if (fprintf(fh, "grep_filter=%s,%s,%s,%s\n", str_or_nothing(gf_list[loop].sserver), str_or_nothing(gf_list[loop].schannel), str_or_nothing(gf_list[loop].file), gf_list[loop].re) <= 0)
			return FALSE;
	}

	return TRUE;
}
